/*
** iplog_filter.c - iplog traffic filter routines.
** Copyright (C) 1999 Odin <odin@numb.org>, behe <eric@ojnk.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <config.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <netdb.h>
#include <iplog.h>

#include <icmp-compat.h>
#include <iplog_filter.h>

static int get_icmp_code(const u_char *);
static int get_addr(const u_char *, u_long *);
static int get_port(const u_char *, int, const u_char *);
static int get_portrange(const u_char *, struct port_range *, const u_char *);
static bool port_match(const struct port_range *, u_short);
static bool addr_match(const struct addr_entry *, u_long);
static size_t make_pretty(const u_char *buf, u_char *dest, size_t len);

static const u_char *proto_list[] = { "tcp", "udp",	"icmp" };
static u_char reason[64];
static struct filter_data *filters[3];

bool apply_icmp_filter(u_int prot, const struct ip *ip, u_char type)
{
	const struct filter_data *rule, *head = filters[prot];

	for (rule = head ; rule != NULL ; rule = rule->next) {
		if (addr_match(&rule->src, ip->ip_src.s_addr) &&
			addr_match(&rule->dst, ip->ip_dst.s_addr) &&
			(rule->dport.min == type || rule->dport.min == ICMP_UNDEFINED))

			return (false ^ rule->not);
	}

	return (false);
}

bool apply_filter(u_int p, const struct ip *ip, u_short sport, u_short dport) {
	const struct filter_data *rule, *head = filters[p];

	for (rule = head ; rule != NULL ; rule = rule->next) {
		if (addr_match(&rule->src, ip->ip_src.s_addr) &&
			addr_match(&rule->dst, ip->ip_dst.s_addr) &&
			port_match(&rule->dport, htons(dport)) &&
			port_match(&rule->sport, htons(sport)))

			return (false ^ rule->not);
	}

	return (false);
}

static bool addr_match(const struct addr_entry *test, u_long host) {
	if (test->addr == 0 || host == 0 || test->addr == host)
		return (true ^ test->not);

	if (((host ^ test->addr) & test->mask) == 0)
		return (true ^ test->not);

	return (false ^ test->not);
}

static bool port_match(const struct port_range *test, u_short port) {
	if (!test->max && !test->min)
		return (true ^ test->not);

	if (test->max == test->min)
		return ((test->max == port) ^ test->not);

	if (!test->max)
		return ((test->min <= port) ^ test->not);
	else if (!test->min)
		return ((test->max >= port) ^ test->not);
	else
		return (((test->max >= port) && (test->min <= port)) ^ test->not);

	return (false ^ test->not);
}

static int get_addr(const u_char *hostname, u_long *addr) {
	struct in_addr in;

	if (inet_aton(hostname, &in)) {
		*addr = in.s_addr;
		return (0);
	} else {
		struct hostent *host = gethostbyname(hostname);
		if (host != NULL) {
			*addr = *((u_long *) host->h_addr);
			return (0);
		}
	}

	return (-1);
}

static int get_icmp_code(const u_char *ptr) {
	if (valid_number(ptr)) {
		int type = strtol(ptr, NULL, 10);

		if (type < 0 || type > 18)
			return (-1);
		return (type);
	}

	if (!strncasecmp(ptr, "ICMP_", 5))
		ptr += 5;

	if (!strncasecmp(ptr, "ECHOREPLY", 5))
		return (ICMP_ECHO_REPLY);
	else if (!strncasecmp(ptr, "ECHO", 2))
		return (ICMP_ECHO);
	else if (!strncasecmp(ptr, "UNREACH", 4))
		return (ICMP_DEST_UNREACHABLE);
	else if (!strncasecmp(ptr, "REDIRECT", 3))
		return (ICMP_REDIRECT);
	else if (!strncasecmp(ptr, "ROUTERADVERT", 9))
		return (ICMP_ROUTER_ADVERT);
	else if (!strncasecmp(ptr, "ROUTERSOLICIT", 9))
		return (ICMP_ROUTER_SOLICIT);
	else if (!strncasecmp(ptr, "TIMXCEED", 5))
		return (ICMP_TIME_EXCEEDED);
	else if (!strncasecmp(ptr, "TIMESTAMP_REPLY", 10))
		return (ICMP_TIMESTAMP_REPLY);
	else if (!strncasecmp(ptr, "TSTAMP", 6))
		return (ICMP_TIMESTAMP);
	else if (!strncasecmp(ptr, "IREQREPLY", 5))
		return (ICMP_INFO_REPLY);
	else if (!strncasecmp(ptr, "IREQ", 2))
		return (ICMP_INFO_REQUEST);
	else if (!strncasecmp(ptr, "MASKREPLY", 5))
		return (ICMP_ADDRESS_REPLY);
	else if (!strncasecmp(ptr, "MASKREQ", 2))
		return (ICMP_ADDRESS);
	else if (!strncasecmp(ptr, "PARAMPROB", 4))
		return (ICMP_PARAMETER_PROBLEM);
	else if (!strncasecmp(ptr, "SOURCEQUENCH", 3))
		return (ICMP_SOURCE_QUENCH);
	else
		snprintf(reason, sizeof(reason), "Unknown icmp type: %s", ptr);

	return (-1);
}

static int get_port(const u_char *s, int wildcard, const u_char *proto) {
	int temp;

	if (!strcasecmp(s, EVERYTHING))
		temp = wildcard;
	else {
		temp = strtol(s, NULL, 0);
		if (temp == 0) {
			struct servent *se = getservbyname(s, proto);
			if (se == NULL) 
				return (-1);
			temp = se->s_port;
		}
	}

	return (temp);
}

static int get_addrrange(const u_char *b, struct addr_entry *addr) {
	u_char *p, *q;
	int i = 0;

	if (*b == '!') {
		addr->not = true;
		++b;
	} else
		addr->not = false;
	
	p = strchr(b, '/');
	if (p != NULL) {
		*p++ = '\0';
		for (q = p ; *q != '\0' ; q++) {
			if (!isdigit(*q)) {
				i = 1;
				break;
			}
		}
			
		if (i == 1) {
			if (get_addr(p, (u_long *) &addr->mask) == -1) {
				snprintf(reason, sizeof(reason), "Bad mask: %s", b);
				return (-1);
			}
		} else {
			if (!strcasecmp(p, EVERYTHING))
				addr->mask = 1;
			else {
				i = strtoul(p, NULL, 0);
				addr->mask = (1 << i) - 1;
			}
		}
	} else
		addr->mask = MASK_MATCH_ONE;

	if (!strcasecmp(b, EVERYTHING))
		addr->addr = 0;
	else if (get_addr(b, &addr->addr) == -1) {
		snprintf(reason, sizeof(reason), "No such host: %s", b);
		return (-1);
	}
	
	return (0);
}

static int get_portrange(const u_char *b, struct port_range *port, const u_char *proto) {
	u_char *p;

	if (*b == '!') {
		port->not = true;
		++b;
	} else
		port->not = false;

	p = strchr(b, ':');
	if (p != NULL)
		*p++ = '\0';

	port->min = get_port(b, MIN_PORT, proto);

	if (p != NULL) {
		/* Allow for stuff like 6000: */
		if (*p == '\0')
			port->max = MAX_PORT;
		else
			port->max = get_port(p, MAX_PORT, proto);
	} else 
		port->max = port->min;

	if ((port->min & MAX_PORT) != port->min ||
		(port->max & MAX_PORT) != port->max)
	{
		snprintf(reason, sizeof(reason),
			"Error: Ports must be between 1 and 65535");
		return (-1);
	}

	return (0);
}

void read_filters(const u_char *filename) {
	FILE *fp;
	bool tty = isatty(STDERR_FILENO);
	u_int i, z;
	struct filter_data data;
	u_char *p = NULL, *q = NULL, *b = NULL, line = 0, let = 0;
	u_char buf[1024], fbuf[1024], proto = 0;
	void *(*atl)(const void *, void **, size_t) = NULL;

	fp = fopen(filename, "r");

	if (fp == NULL)
		return;

	goto begin;

ojnk:
	if (let) {
		if (reason[0] == '\0')
			xstrncpy(reason, "format error", sizeof(reason));

		if (tty == true)
			fprintf(stderr, "%s:%d: %s\n", filename, line, reason);
		else
			mysyslog(PRIORITY, "%s:%d: %s", filename, line, reason);
	}
	reason[0] = '\0';

begin:
	while (get_line(fp, buf, sizeof(buf) - 1) != EOF) {
		++line;

		let = make_pretty(buf, fbuf, sizeof(fbuf));
		if (let == 0)
			continue;

		data.sport.min = MIN_PORT;
		data.sport.max = MAX_PORT;
		data.sport.not = false;
		data.dport.max = MAX_PORT;
		data.dport.not = false;
		data.src.addr = ADDR_MATCH_ALL;
		data.src.mask = MASK_MATCH_ONE;
		data.src.not = false;
		data.dst.addr = ADDR_MATCH_ALL;
		data.dst.mask = MASK_MATCH_ONE;
		data.dst.not = false;
		data.not = false;

		for (p = fbuf, z = 0, i = 0 ; p != NULL ;) {
			q = p;
			p = strchr(q, ' ');
			if (p == NULL)
				break;
			*p++ = '\0';

			if (i == 0) {
				if (!strcasecmp(q, "log")) {
					atl = list_prepend;
					data.not = false;
				} else if (!strcasecmp(q, "ignore")) {
					atl = list_append;
					data.not = true;
				} else {
					snprintf(reason, sizeof(reason),
						"First field must be either \"log\" or \"ignore\". "
						"  You gave \"%s\".", q);
					goto ojnk;
				}
				++i;
				continue;
			}
			
			if (i == 1) {
				if (!strcasecmp(q, "tcp"))
					proto = FIL_TCP;
				else if (!strcasecmp(q, "udp"))
					proto = FIL_UDP;
				else if (!strcasecmp(q, "icmp")) {
					data.dport.min = ICMP_UNDEFINED;
					proto = FIL_ICMP;
				} else {
					snprintf(reason, sizeof(reason),
						"Bad protocol: %s", q);
					goto ojnk;
				}
				++i;
				continue;
			}

			b = q;
			q = p;
			p = strchr(q, ' ');
			if (p != NULL)
				*p++ = '\0';
			if (proto != FIL_ICMP) {
				if (!strcasecmp(b, "dport"))
					get_portrange(q, &data.dport, proto_list[proto]);
				else if (!strcasecmp(b, "sport"))
					get_portrange(q, &data.sport, proto_list[proto]);
			} else if (!strcasecmp(b, "type")) {
				data.dport.min = get_icmp_code(q);
				if (data.dport.min == -1)
					goto ojnk;
			}
			if (!strcasecmp(b, "from"))
				get_addrrange(q, &data.src);
			else if (!strcasecmp(b, "to"))
				get_addrrange(q, &data.dst);
			else if (strcasecmp(b, "dport") && strcasecmp(b, "sport") &&
					strcasecmp(b, "type"))
			{
				snprintf(reason, sizeof(reason), "Unknown keyword: \"%s\"", b);
				goto ojnk;
			}
		}

		atl(&data, (void **) &filters[proto], sizeof(data));
	}

	fclose(fp);
}

static size_t make_pretty(const u_char *buf, u_char *dest, size_t len) {
	size_t i = 0;
	u_char c, l = 0;

	while (isspace(*buf))
		++buf;

	while (i < len && *buf != '\0' && *buf != '#' && *buf != '\n') {
		c = *buf++;
		switch (c) {
			case '\t':
				c = ' ';
				/* FALLTHROUGH */
			case ' ':
				if (i == 0 || l == ' ')
					break;
				/* FALLTHROUGH */
			default:
				dest[i++] = l = c;
		}
	}

	if (i > 0 && dest[i - 1] == ' ')
		dest[i - 1] = '\0';

	dest[i] = '\0';
	return (i);
}

void add_dns_ignore_rules(void) {
	struct filter_data temp;
	u_char buf[1024], nbuf[1024], *p;
	FILE *fp;

	fp = fopen("/etc/resolv.conf", "r");
	if (fp == NULL)
		return;

	temp.sport.min = 53;
	temp.sport.max = 53;
	temp.dport.min = MIN_PORT;
	temp.dport.max = MAX_PORT;
	temp.src.not = false;
	temp.src.mask = MASK_MATCH_ONE;
	temp.dst.addr = ADDR_MATCH_ALL;
	temp.dst.mask = MASK_MATCH_ONE;
	temp.dst.not = false;
	temp.not = true;
	
	while (get_line(fp, buf, sizeof(buf) - 1) != EOF) {
		if (make_pretty(buf, nbuf, sizeof(nbuf)) == 0)
			continue;

		p = strchr(nbuf, ' ');
		if (p == NULL)
			break;

		get_addr(++p, &temp.src.addr);
		list_prepend(&temp, (void **) &filters[FIL_UDP], sizeof(temp));
	}

	fclose(fp);
}

#ifdef HAVE_PTHREAD_CANCEL
void destroy_filter_list(u_int prot) {
	list_destroy(filters[prot], NULL);
	filters[prot] = NULL;
}
#endif

/* vim:ts=4:sw=8:tw=0 */
